# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import numpy as np
from datetime import datetime

from quocslib.stoppingcriteria.StoppingCriteria import StoppingCriteria


class CMAESStoppingCriteria(StoppingCriteria):
    terminate_reason: str
    is_converged: bool

    def __init__(self, stopping_criteria: dict):
        """
        Class for the Nelder Mead custom stopping criteria
        :param dict stopping_criteria:
        """
        # Call to the super class constructor
        super().__init__(stopping_criteria)
        # frtol and xatol
        self.xatol = stopping_criteria.setdefault("xatol", 1e-14)
        self.frtol = stopping_criteria.setdefault("frtol", 1e-13)
        self.is_converged = False
        self.terminate_reason = ""

    def check_stopping_criteria(self, fsim: np.array = None, func_evaluations_single_direct_search: int = None) -> None:
        """
        :param f_sim:
        :param func_evaluations_single_direct_search:
        :return:
        """
        if self.is_converged: return

        # # Check function evaluation
        # is_converged, terminate_reason = self.check_func_eval(function_evaluations)
        # if is_converged:
        #     self.is_converged = True
        #     self.terminate_reason = terminate_reason
        #     return

        # # Convergence FoM criterion
        # is_converged, terminate_reason = self.check_f_size(f_sim)
        # if is_converged:
        #     self.is_converged = True
        #     self.terminate_reason = terminate_reason
        #     return

        # self.is_converged, self.terminate_reason = self.check_func_eval_total(func_evaluations_single_direct_search)
        # if self.is_converged: return

        self.is_converged, self.terminate_reason = self.check_func_eval_single_direct_search(func_evaluations_single_direct_search)
        if self.is_converged: return

        self.is_converged, self.terminate_reason = self.check_f_size(fsim)
        if self.is_converged: return

        # self.is_converged, self.terminate_reason = self.check_goal_reached(fsim[0])
        # if self.is_converged: return

        # self.is_converged, self.terminate_reason = self.check_total_time_out()
        # if self.is_converged: return

        self.is_converged, self.terminate_reason = self.check_direct_search_time_out()
        if self.is_converged: return



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import os
import time
import numpy as np

from quocslib.utils.AbstractDump import AbstractDump


class BestDump(AbstractDump):
    def __init__(self, results_path: str = ".", date_time: str = ".", **kwargs):
        """
        Dumping class for controls and other data which should be the most useful option for most users.
        :param str results_path: Path of the folder of the results
        :param str date_time: String containing the identifier in the form of date and time
        """
        self.best_controls_path = results_path
        self.date_time = date_time

    def dump_controls(self,
                      pulses: list = [],
                      timegrids: list = [],
                      parameters: list = [],
                      is_record: bool = False,
                      **kwargs) -> None:
        """
        Save the controls in the results folder
        :param list pulses: the list containing the pulses that were optimized
        :param list timegrids: the list containing the time grids that were used in the optimization
        :param list parameters: the list containing the parameters that were optimized
        :param bool is_record: information if the current controls are a new record
        """
        if not is_record:
            return

        controls_dict = {}
        pulse_index = 1

        for pulse, time_grid in zip(pulses, timegrids):
            controls_dict["pulse" + str(pulse_index)] = pulse
            controls_dict["time_grid" + str(pulse_index)] = time_grid
            pulse_index += 1

        parameter_index = 1

        for parameter in parameters:
            controls_dict["parameter" + str(parameter_index)] = parameter
            parameter_index += 1

        # Full dictionary
        full_dict = {**controls_dict, **kwargs}

        # Save the file
        controls_path = os.path.join(self.best_controls_path, self.date_time + "_best_controls.npz")
        np.savez(controls_path, **full_dict)

    def other_dumps(self, filename: str = "test.txt", data: np.array = np.array([0.0])):
        """
        Save other results into a txt numpy file
        :param str: filename
        :param np.array: data
        """
        # Create the path
        path = os.path.join(self.best_controls_path, filename)
        # Save the data in a txt file
        np.savetxt(path, data)



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from quocslib.utils.AbstractFoM import AbstractFoM
from quocslib.utils.inputoutput import writejsonfile
import numpy as np
import os
import time


class FilesUpdateFoM(AbstractFoM):
    """An evaluation method for the figure of merit via files exchange. The communication object accesses to the
    get_FoM function.
    The get_FoM removes the "FoM.txt" file and creates a json, a txt or a npz file in the controls
    folder designed by the user.
    Finally, the read_FoM_values will wait and check for the figure of merit in the folder designed by the user.
    In case none figure of merit is provided by the user in the limited time defined by the user in the constructor or
    an error occur during the evaluation an error will set in the
    """
    def __init__(self,
                 controls_folder: str = ".",
                 is_splitted: bool = False,
                 file_extension: str = "json",
                 FoM_folder: str = ".",
                 max_time: float = 60 * 2,
                 **kwargs) -> None:
        """

        :param str controls_folder: Path of the controls folder
        :param str FoM_folder: Path of the figure of merit folder
        :param int max_time: Maximum time in arbitrary units. 1 = 0.5 seconds
        :param kwargs: Other parameters
        """
        # Control folder
        self.controls_path = os.path.join(controls_folder, "controls.{0}".format(file_extension))
        # File extension
        self.file_extension = file_extension
        # Split the controls in multiple files
        self.is_splitted = is_splitted
        # FoM folder
        self.FoM_path = os.path.join(FoM_folder, "FoM.txt")
        # Maximum time in seconds to wait for the figure of merit evaluation
        self.max_time = max_time

    def get_FoM(self, pulses: list = [], timegrids: list = [], parameters: list = []) -> dict:
        """
        Write the controls in the controls.npz file, read the figure of merit in the FoM.txt file
        """
        print("Removing the previous {0} file if any".format(self.FoM_path))
        # If FoM.txt file exists remove it
        if os.path.exists(self.FoM_path):
            os.remove(self.FoM_path)
        # Write the pulses into a file or multiple files
        # TODO Multiple files extension
        print("Putting controls in {0}".format(self.controls_path))
        self.put_controls_into_user_path(pulses, timegrids, parameters)
        # Read the content of FoM.txt file
        print("Reading the {0} file ".format(self.FoM_path))
        return self.read_FoM_values()

    def read_FoM_values(self) -> dict:
        """
        Read the figure of merit FoM.txt file
        :return: dict The figure of merit dictionary
        """
        # Read the FoM
        FoM = None
        time_counter = 0
        time_counter_max = self.max_time * 2
        # Check if figure of merit file exists
        while not os.path.exists(self.FoM_path):
            time_counter += 1
            time.sleep(0.5)
            if time_counter >= time_counter_max:
                return {"FoM": FoM, "status_code": -2}
        # Sleep to be sure the file is correctly close
        time.sleep(0.01)
        try:
            with open(self.FoM_path, "r") as FoM_file:
                FoM = float(str(FoM_file.readline()).strip())
        # TODO Add specific exception why the figure of merit is not readable
        except Exception as ex:
            print("Unhandled exception during FoM reading: {0}".format(ex.args))
            return {"FoM": FoM, "status_code": -3}
        return {"FoM": FoM}

    def put_controls_into_user_path(self, pulses_list: list, time_grids_list: list, parameters_list: list) -> None:
        """
        Save the controls in the controls.npz file
        :param list pulses_list: List of np.arrays. One np.array for each pulse
        :param list time_grids_list: List of np.arrays. One np.array for each time grid
        :param list parameters_list: List of floats. One float ofr each parameter
        :return:
        """
        # Choose between save as json file or txt file
        file_extension = self.file_extension
        if file_extension == "txt":
            self._put_controls_txt(pulses_list, time_grids_list, parameters_list)
        elif file_extension == "json":
            self._put_controls_json(pulses_list, time_grids_list, parameters_list)
        else:
            # TODO Return a negative status code
            print("The extension {0} is not recognized".format(file_extension))

    def _put_controls_txt(self, pulses_list: list, time_grids_list: list, parameters_list: list):
        """Ordered the controls like pulse1, time_grid1, pulse2, time_grid2, ... para1, para2 ... and save into a
        txt file.
        """
        with open(self.controls_path, "wb") as controls_file:
            for pulse, time_grid in zip(pulses_list, time_grids_list):
                np.savetxt(controls_file, [pulse], fmt="%f")
                np.savetxt(controls_file, [time_grid], fmt="%f")

            np.savetxt(controls_file, parameters_list, fmt="%f")

    def _put_controls_json(self, pulses_list: list, time_grids_list: list, parameters_list: list):
        """Save into a json file"""
        # Save the pulses and the respective timegrids into a dictionary
        controls_dict = {}
        pulse_index = 1
        # The pulses are saved like pulse#
        # The timegrids are saved like time_grids#
        for pulse, time_grid in zip(pulses_list, time_grids_list):
            controls_dict["pulse" + str(pulse_index)] = pulse
            controls_dict["time_grid" + str(pulse_index)] = time_grid
            pulse_index += 1
        parameter_index = 1
        # The parameters are saved like parameter#
        for parameter in parameters_list:
            controls_dict["parameter" + str(parameter_index)] = parameter
            parameter_index += 1
        writejsonfile(self.controls_path, controls_dict)



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
from ast import Import

try:
    import jax.numpy as jnp
    import jax.scipy as jsp
except:
    raise ImportError

from quocslib.optimizationalgorithms.OptimizationAlgorithm import OptimizationAlgorithm
from quocslib.Controls import Controls
from quocslib.timeevolution.piecewise_integrator import pw_final_evolution


class ADAlgorithm(OptimizationAlgorithm):
    """
    This is the template for an algorithm class. The three important function are:
    * the constructor with the optimization dictionary and the communication object as parameters
    * run : The main loop for optimal control
    * _get_response_for_client : return info about the goodness of the controls and errors if any
    * _get_controls : return the set of controls as a dictionary with pulses, parameters, and times as keys
    * _get_final_results: return the final result of the optimization algorithm
    """
    def __init__(self, optimization_dict: dict = None, communication_obj=None):
        """
        This is the implementation of the GRAPE algorithm. All the arguments in the constructor are passed to the
        OptimizationAlgorithm class except the optimization dictionary where the GRAPE settings and the controls are defined.
        """
        super().__init__(communication_obj=communication_obj)
        ###########################################################################################
        # Optimal algorithm variables if any
        ###########################################################################################
        alg_parameters = optimization_dict["algorithm_settings"]
        # Starting FoM
        # self.best_FoM = 1e10  # defined in parent class
        ###########################################################################################
        # Pulses, Parameters, Times object
        ###########################################################################################
        # Initialize the control object
        self.controls = Controls(
            optimization_dict["pulses"],
            optimization_dict["times"],
            optimization_dict["parameters"],
        )

        # might need to control if you change something

        self.A = optimization_dict["A"]
        self.B = optimization_dict["B"]
        self.n_slices = optimization_dict["n_slices"]
        self.rho_init = optimization_dict["rho_init"]
        self.rho_target = optimization_dict["rho_target"]
        self.dt = optimization_dict["dt"]
        self.sys_type = optimization_dict["sys_type"]
        self.dim = jnp.size(self.A, 1)
        self.u0 = optimization_dict["u0"]

        self.optimized_pulses = None
        self.opt_res = None

    def functional(self, drive, A, B, n_slices, dt, u0, rho0, rhoT, sys_type):
        """Compute the fidelity functional for the given problem defined in the class

        :param jnp.array drive: a flat array that contains the pulse amplitudes
        :param jnp.matrix A: the drift hamiltonian
        :param List[jnp.matrix] B: the control hamiltonians
        :param int n_slices: the number of slices in the pulse
        :param float dt: the duration of each timeslice
        :param jnp.matrix u0: the initial propagator that should be used
        :param jnp.matrix rho0: the initial density matrix
        :param jnp.matrix rhoT: the target density matrix
        :param str sys_type: the string to indicate the system type, can be either StateTransfer or left blank
        :return float: the value of the fidelity at the current point in time
        """
        K = len(B)
        drive = drive.reshape((K, n_slices))
        U = pw_final_evolution(drive, A, B, n_slices, dt, u0)

        if sys_type == "StateTransfer":
            ev = U @ rho0 @ U.T.conj()
        else:
            ev = U @ rho0
        fid = 1 - jnp.abs(jnp.trace(ev @ rhoT.T.conj()))

        return fid

    def _get_functional(self):
        """generates a lambda function lambda x: which evaluates and returns the figure of merit

        :return lambda:
        """
        return lambda x: self.functional(
            x,
            self.A,
            self.B,
            self.n_slices,
            self.dt,
            self.u0,
            self.rho_init,
            self.rho_target,
            self.sys_type,
        )

    def run(self) -> None:
        """This runs the main loop of the optimization, assuming that everything
        has been configured correctly this should use LBFGS, or a chosen algorithm,
        to optimize the pulse
        """

        func_topt = self._get_functional()
        init = self.controls
        # now we can optimize
        # need to be able to include things
        optimization_result = jsp.optimize.minimize(func_topt, init, method="BFGS")

        # need to be able to implement pulses in Marco's way, ask him later
        self.best_FoM = optimization_result.fun
        self.optimized_pulses = optimization_result.x
        self.opt_res = optimization_result

        #     # Update the base current pulses
        #     self._update_base_pulses()

    # def _update_base_pulses(self) -> None:
    #     """Update the base dCRAB pulse"""
    #     self.controls.update_base_controls(self.xx)

    def _get_controls(self, optimized_control_parameters: jnp.array) -> dict:
        """Get the controls dictionary from the optimized control parameters

        :param jnp.array optimized_control_parameters: the array of optimize control parameters
        :return dict: returns a dict that contains the pulses, parameters and timegrid
        """
        [pulses, timegrids, parameters] = self.controls.get_controls_lists(optimized_control_parameters)
        #
        controls_dict = {
            "pulses": pulses,
            "parameters": parameters,
            "timegrids": timegrids,
        }
        return controls_dict

    def _get_final_results(self) -> dict:
        """Return a dictionary with final results to put into a dictionary"""
        final_dict = {
            "Figure of merit": self.best_FoM,
            "total number of function evaluations": self.iteration_number,
        }
        return final_dict

    def _get_response_for_client(self) -> dict:
        raise NotImplementedError



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import logging
import os
import sys


def create_logger(results_path: str, logger_name: str = "oc_logger", is_debug: bool = False):
    """Logger creation for console, log file, and debug log file"""
    log_format = "%(asctime)s.%(msecs)03d - %(name)s - %(levelname)s - %(message)s"
    date_format = "%m/%d/%Y %I:%M:%S"
    print_format = "%(levelname)-8s %(name)-12s %(message)s"
    log_filename = os.path.join(results_path, "logging.log")
    log_debug_filename = os.path.join(results_path, "logging_debug.log")

    logger = logging.getLogger(logger_name)
    # Remove previous handlers if any
    logger.handlers = []
    # Default level for logger
    logger.setLevel(logging.DEBUG)
    # Console handler
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(logging.Formatter(print_format))
    # Log file handler
    file_handler = logging.FileHandler(log_filename)
    file_handler.setLevel(logging.INFO)
    file_handler.setFormatter(logging.Formatter(log_format, date_format))
    # Add handlers to the logger
    logger.addHandler(file_handler)
    logger.addHandler(console_handler)
    ################################################
    # Debug
    # In case of debug another object handler is activated
    ################################################
    if is_debug:
        # Log debug file handler
        debug_file_handler = logging.FileHandler(log_debug_filename)
        debug_file_handler.setLevel(logging.DEBUG)
        debug_file_handler.setFormatter(logging.Formatter(log_format, date_format))
        #
        logger.addHandler(debug_file_handler)
    return logger



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from quocslib.optimalcontrolproblems.su2 import hamiltonian_d1_d2_2fields
import numpy as np
from scipy.linalg import expm, norm
from quocslib.utils.AbstractFoM import AbstractFoM


class OneQubit2Fields(AbstractFoM):
    def __init__(self, args_dict: dict = None):
        if args_dict is None:
            args_dict = {}

        self.psi_target = np.asarray(
            eval(args_dict.setdefault("target_state", "[1.0/np.sqrt(2), -1j/np.sqrt(2)]")),
            dtype="complex",
        )
        self.psi_0 = np.asarray(eval(args_dict.setdefault("initial_state", "[1.0, 0.0]")), dtype="complex")
        self.delta1 = args_dict.setdefault("delta1", 0.1)
        self.delta2 = args_dict.setdefault("delta2", 0.1)
        # Noise in the figure of merit
        self.is_noisy = args_dict.setdefault("is_noisy", False)
        self.noise_factor = args_dict.setdefault("noise_factor", 0.05)
        self.std_factor = args_dict.setdefault("std_factor", 0.01)

    def get_FoM(self, pulses: list = [], parameters: list = [], timegrids: list = []) -> dict:
        amplitude = np.asarray(pulses[0])
        phase = np.asarray(pulses[1])
        timegrid = np.asarray(timegrids[0])
        dt = timegrid[1] - timegrid[0]
        U = self._time_evolution(amplitude, phase, dt, self.delta1, self.delta2)
        psi_f = np.matmul(U, self.psi_0)
        infidelity = 1.0 - self._get_fidelity(self.psi_target, psi_f)

        std = 0.0
        if self.is_noisy:
            noise = (self.noise_factor * 2 * (0.5 - np.random.rand(1, )[0]))
            infidelity += noise
            std = (self.std_factor * np.random.rand(1, )[0])

        return {"FoM": np.abs(infidelity), "std": std}

    @staticmethod
    def _time_evolution(amplitude, phase, dt, delta1=0.0, delta2=0.0):
        U = np.identity(2)
        for ii in range(amplitude.size - 1):
            ham_t = hamiltonian_d1_d2_2fields(
                (amplitude[ii + 1] + amplitude[ii]) / 2,
                (phase[ii + 1] + phase[ii]) / 2,
                delta1,
                delta2,
            )
            U_temp = U
            U = np.matmul(expm(-1j * ham_t * dt), U_temp)
        return U

    @staticmethod
    def _get_fidelity(psi1, psi2):
        return np.abs(np.dot(psi1.conj().T, psi2))**2 / (norm(psi1) * norm(psi2))



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import numpy as np
from quocslib.tools.randomgenerator import RandomNumberGenerator


def ptrace(rho, dimensions):
    """
    Useful to have this implementation of the partial trace which uses einsums

    TODO implement this in Python again
    """

    return rho


def commutator(A, B):
    return A @ B - B @ A


def gram_schmidt(A):
    """
    Orthonormalize a set of linear independent vectors

    :param A: Square matrix with linear independent vectors
    :return A: Square matrix with orthonormalize vectors
    """
    # Get the number of vectors.
    n = A.shape[1]
    for j in range(n):
        # To orthogonalize the vector in column j with respect to the
        # previous vectors, subtract from it its projection onto
        # each of the previous vectors.
        for k in range(j):
            u_k = A[:, k]
            A[:, j] -= np.dot(u_k, A[:, j]) * u_k / np.linalg.norm(u_k)**2
        A[:, j] = A[:, j] / np.linalg.norm(A[:, j])
    return A


def simplex_creation(mean_value: np.array, sigma_variation: np.array, rng: RandomNumberGenerator = None) -> np.array:
    """
    Creation of the simplex

    @return:
    """
    ctrl_par_number = mean_value.shape[0]
    ##################
    # Scale matrix:
    # Explain what the scale matrix means here
    ##################
    # First row
    x0_scale = np.zeros((1, ctrl_par_number))
    # Simplex matrix ( without first row )
    simplex_matrix = np.diag(np.ones_like(sigma_variation))
    # Add random number in the first column
    if rng is None:
        random_array = np.random.rand(ctrl_par_number)
    else:
        random_array = rng.get_random_numbers(ctrl_par_number)
    random_array = random_array.reshape(ctrl_par_number, )
    simplex_matrix[0, :] += np.sqrt(3) * (random_array - 0.5) * 2
    # Orthogonalize set of vectors with gram_schmidt, and rescale with the normalization length
    simplex_matrix_orthonormal = gram_schmidt(simplex_matrix.T)
    # Rescale the vector with the sigma variation
    simplex_matrix_orthogonal_rescaled = simplex_matrix_orthonormal @ np.diag(sigma_variation)
    # Add the first row containing only zeros
    x_t_norm = np.append(x0_scale, simplex_matrix_orthogonal_rescaled, axis=0)
    # Offset matrix
    x_offset = np.outer(np.ones((1, ctrl_par_number + 1)), mean_value)
    # Start simplex matrix
    StartSimplex = x_t_norm + x_offset
    return StartSimplex


if __name__ == "__main__":
    # TODO Move this main script to a test script
    Nc = 4
    ampl_var_1 = 2.0
    ampl_var_2 = 0.7
    f_norm = 1 / np.sqrt(2)
    p_1 = (ampl_var_1 * f_norm) * np.ones(2, )
    p_2 = (ampl_var_2 * f_norm) * np.ones(2, )
    sc_vec = np.append(p_1, p_2)

    x0_scale = np.zeros((1, Nc))
    # Simplex matrix ( without first row )
    simplex_m = np.diag(sc_vec)
    # Add random number
    simplex_m[0, :] += ((sc_vec[0] / 10.0) * (np.random.rand(Nc, ) - 0.5) * 2)

    simplex_m_r = gram_schmidt(simplex_m.T, sc_vec).T
    # Rescale accordingly to amplitude variation
    # x_norm = A_norm.dot(np.diag(sc_vec))
    # Add first row
    x_t_norm = np.append(x0_scale, simplex_m_r, axis=0)

    print(x_t_norm)


def to_sup_op(H):
    """
    Function to convert a Hamiltonian into a Liouvillian
    """
    dim = np.size(H, 1)
    idm = np.eye(dim)
    return np.kron(idm, H) - np.kron(H.T.conj(), idm)


def to_vec(rho):
    """
    Take an input rho vector and flatten it into a column
    """
    return rho.flatten()



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import numpy as np
from packaging import version


class RandomNumberGenerator:
    def __init__(self, seed_number: int = None):
        numpy_version = np.__version__
        self.message = ""
        self.rng = None
        self.type = None
        # If no seed is provided use the random numbers without any seed
        if seed_number is None:
            return
        if version.parse(numpy_version) < version.parse("1.16.0"):
            # Try to import the randomgen package
            try:
                import randomgen
                from randomgen import RandomGenerator, MT19937

                self.message = "Import the randomgen library version: {}".format(randomgen.__version__)
                self.rng = RandomGenerator(MT19937(seed=seed_number))
                self.type = "randomgen"
            except ImportError:
                raise ImportError(
                    "Please install randomgen using a compatible version of numpy {0}".format(numpy_version))
        else:
            self.rng = np.random.default_rng(seed_number)
            self.type = "numpy"

    def get_random_numbers(self, n: int):
        """Return an array of random numbers"""
        if self.rng is None:
            return np.random.rand(n)
        else:
            if self.type == "numpy":
                return self.rng.random(n)
            else:
                return self.rng.random_sample(n)


# def get_random_numbers(n: int, rng: np.random.Generator = None):
#     """ Return an array of random numbers between 0 and 1 based on the random generator """
#     if rng is None:
#         return np.random.rand(n)
#     else:
#         return rng.random(n)



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import numpy as np

# np.seterr(all="raise")

from scipy import linalg

from quocslib.gradientfreemethods.DirectSearchMethod import DirectSearchMethod
from quocslib.stoppingcriteria.CMAESStoppingCriteria import CMAESStoppingCriteria


class CMAES(DirectSearchMethod):
    callback: callable

    def __init__(self, settings: dict = {}, stopping_criteria: dict = {}, callback: callable = None, **kwargs):
        """
        The Covariance matrix adaptation evolution strategy is an updating algorithm based on repeatedly testing
        distributions of points in the control landscape
        :param dict settings: settings for the CMAES algorithm
        :param dict stopping_criteria: stopping criteria
        """
        super().__init__()
        self.callback = callback
        # Active the parallelization for the firsts evaluations
        self.is_parallelized = settings.setdefault("parallelization", False)
        self.is_adaptive = settings.setdefault("is_adaptive", False)
        # TODO Create it using dynamical import module
        # Stopping criteria object
        self.sc_obj = CMAESStoppingCriteria(stopping_criteria)

    def run_dsm(self,
                func,
                x0,
                args=(),
                sigma_v: np.array = None,
                initial_simplex=None,
                drift_comp_minutes=0.0,
                **kwargs) -> dict:
        """

        :param callable func: Function tbe called at every function evaluation
        :param np.array x0: initial point
        :param tuple args: Further arguments
        :param np.array initial_simplex: Starting simplex for the Nelder Mead evaluation
        :param float drift_comp_minutes: Compensate for drift after this number of minutes
        :return:
        """

        # Creation of the communication function for the OptimizationAlgorithm object
        calls_number, func = self._get_wrapper(args, func)

        # Set to false is_converged
        self.sc_obj.is_converged = False

        N = len(x0)

        xmean = x0
        # Sigma
        sigma = 1.0
        # coordinate wise standard deviation (step-size) TR 2020_04_15: use ReasonableAmplVar (see later in for loop)
        if sigma_v is None or len(sigma_v) != N:
            sigma_v = 0.3 * np.ones(N, )

        # Strategy parameter setting: Selection population size, offspring number TR 2020_04_15: according to The CMA
        # Evolution Strategy: A Tutorial (Hansen) this can be increased number of parents/points for recombination
        l_pop = int(4 + np.floor(3 * np.log(N)))
        mu = int(np.floor(l_pop / 2))
        # muXone array for weighted recombination
        weights = np.log(mu + 0.5) - np.log(np.linspace(1, mu, num=mu))
        # Normalize recombination weights array
        weights = weights / np.sum(weights)
        # variance-effectiveness
        mueff = 1 / sum(weights**2)

        # Strategy parameter setting: Adaptation Time constant for cumulation for C
        cc = (4 + mueff / N) / (N + 4 + 2 * mueff / N)
        # t-const for cumulation fror sigma control
        cs = (mueff + 2) / (N + mueff + 5)
        # learning rate for rank-one update of C
        c1 = 2 / ((N + 1.3)**2 + mueff)
        # learning rate for rank-mu update of C
        cmu = np.minimum(1 - c1, 2 * (mueff - 2 + 1 / mueff) / ((N + 2)**2 + mueff))
        # damping for sigma
        damps = 1 + 2 * np.maximum(0, np.sqrt((mueff - 1) / (N + 1)) - 1) + cs

        # Initialize dynamic (internal) strategy parameters and constants
        pc = np.zeros((N, ))
        ps = np.zeros((N, ))
        B = np.eye(N)
        # Build initial D matrix with scale vector
        D = sigma_v

        # Initial covariance matrix
        C = B * np.diag(D**2) * B.T
        invsqrtC = B * np.diag(D**(-1)) * B.T
        # Eigenvalue approximation
        eigeneval = 0
        # Expectation value
        chiN = N**0.5 * (1 - 1 / (4 * N) + 1 / (21 * N**2))

        # Total evaluation
        counteval = 0
        # Initial simplex size
        arx = np.zeros((N, l_pop))

        # Arguments for stopping criteria
        iterations = 1
        # terminateReason = -1  # JZ 20161125: introduced this quantity

        is_terminated = False

        # figure of merit array
        fsim = np.zeros(l_pop, dtype=float)
        ind = np.zeros(l_pop, dtype=int)

        while not self.sc_obj.is_converged:

            for k in range(l_pop):
                # TR 2020_04_15: Hansen (2016) here also has arz: arz[:,k] = randn(N,)  standard normally
                # distributed vector TR 2020_04_15: here we should make sigma dependent on the reasonable amplitude
                # variation, right?
                arx[:, k] = xmean + sigma * B.dot(D * np.random.randn(N, ))
                # Starting point at the beginning of the SI
                if counteval == 0:
                    arx[:, k] = xmean
                # Possible parallelization here (only for open-loop  optimization)
                fsim[k] = func(arx[:, k], iterations)
                counteval += 1
            iterations = counteval

            # Sort fsim so that lowest value is at 0 and then descending
            ind = np.argsort(fsim)
            fsim = np.take(fsim, ind, 0)

            # Checks general stopping criteria

            # TR 2020_04_15: Does this part make sense here the way it is? For NM I want to consider the Simplex size
            # but here we take the average of the whole population... maybe better use only best value or think about
            # what makes sense here

            if not is_terminated:
                xold = xmean
                # Recombination, new mean value
                xmean = arx[:, ind[0:mu]].dot(weights)
                # TR 2020_04_15: Hansen (2016) here also has zmean:
                # zmean = arz(:, arindex[1:mu]).dot(weights)  # == Dˆ(-1)*B’*(xmean-xold)/sigma
                # New average vector
                y = xmean - xold
                z = invsqrtC.dot(y)

                ps = (1 - cs) * ps + np.sqrt(cs * (2 - cs) * mueff) * z / sigma

                # hsig = np.linalg.norm(ps) / np.sqrt(1 - (1 - cs) ** (2 * counteval / l_pop)) / chiN < 2 + 4. / (N + 1)
                hsig = np.linalg.norm(ps) / np.sqrt(1 - np.power((1 - cs),
                                                                 (2 * counteval / l_pop))) / chiN < 1.4 + 2 / (N + 1)
                # Evolution path update
                pc = (1 - cc) * pc + hsig * np.sqrt(cc * (2 - cc) * mueff) * y / sigma

                # Adapt covariance matrix C, i.e. rank mu update
                artmp = (1 / sigma) * (arx[:, ind[0:mu]] - np.tile(xold, (mu, 1)).T)
                C = ((1 - c1 - cmu) * C + c1 * (np.outer(pc, pc) + (1 - hsig) * cc * (2 - cc) * C) +
                     cmu * artmp.dot(np.diag(weights).dot(artmp.T)))

                # Adapt step size sigma
                sigma = sigma * np.exp((cs / damps) * (np.linalg.norm(ps) / chiN - 1))

                # Decomposition of C into into B*diag(D.^2)*B' (diagonalization) to achieve O(N^2)
                if (counteval - eigeneval) > l_pop / (c1 + cmu) / N / 10:
                    # eigeneval = counteval
                    C = np.triu(C) + np.triu(C, 1).T
                    # (lv, B) = np.linalg.eig(C)
                    (lv, B) = linalg.eigh(C)
                    D = np.sqrt(lv)
                    invsqrtC = B.dot(np.diag(1 / D).dot(B.T))

                if self.callback is not None:
                    if not self.callback():
                        self.sc_obj.is_converged = True
                        self.sc_obj.terminate_reason = "User stopped the optimization or higher order " \
                                                       "stopping criterion has been reached"
                # Check stopping criteria
                self.sc_obj.check_stopping_criteria(fsim, calls_number[0])
                # # CMAES criterium
                # if np.amax(D) > 1e7 * np.amin(D):
                #     FoM_best = fsim[0]
                #     is_terminated = True
                #     terminateReason = "CMAES criterion"  # TR 2020_04_15: Is this description helpful? Shouldn't it be more specific?
                # if iterations > max_eval:
                #     is_terminated = True
                #     terminateReason = "Reached maximum iterations number"

        # Return the best point
        print("Best Result: {0} ,  in {1} evaluations.".format(fsim[0], counteval))
        fval = fsim[0]
        x = arx[:, ind[0]]
        result_custom = {
            "F_min_val": fval,
            "X_opti_vec": x,
            "NitUsed": iterations,
            "NfunevalsUsed": calls_number[0],
            "terminate_reason": self.sc_obj.terminate_reason,
        }

        return result_custom



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from quocslib.utils.AbstractFoM import AbstractFoM
from scipy import optimize
import numpy as np


class Rosenbrock(AbstractFoM):
    """A figure of merit class for optimization of the Rosenbrock function given an arbitrary
    number of parameters"""
    def __init__(self, args_dict: dict = None):
        """Initialize useful arguments"""
        # Noise in the figure of merit
        self.is_noisy = args_dict.setdefault("is_noisy", False)
        self.noise_factor = args_dict.setdefault("noise_factor", 0.05)
        self.std_factor = args_dict.setdefault("std_factor", 0.01)

    def get_FoM(self, pulses: list = [], parameters: list = [], timegrids: list = []) -> dict:
        FoM = optimize.rosen(np.asarray(parameters))
        std = 0.0
        if self.is_noisy:
            noise = (self.noise_factor * 2 * (0.5 - np.random.rand(1, )[0]))
            FoM += noise
            std = (self.std_factor * np.random.rand(1, )[0])

        return {"FoM": FoM, "std": std}



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import pytest
import os
import time
import numpy as np
from quocslib.utils.BestDump import BestDump


def test_dump_controls_record():

    folder = os.path.dirname(os.path.realpath(__file__))
    dump_obj = BestDump(folder, "000")

    test_pulse = np.array([1, 2, 3])
    test_timegrid = np.array([4, 5, 6])
    test_params = np.array([7, 8, 9])

    dump_obj.dump_controls([test_pulse], [test_timegrid], [test_params], True)

    outfile_path = os.path.join(dump_obj.best_controls_path, "000_best_controls.npz")
    controls = np.load(outfile_path)
    # print(controls.files)
    with np.load(outfile_path) as controls:
        assert (controls["pulse1"] == test_pulse).all()
        assert (controls["time_grid1"] == test_timegrid).all()
        assert (controls["parameter1"] == test_params).all()
    os.remove(outfile_path)


def test_dump_controls_NO_record():

    folder = os.path.dirname(os.path.realpath(__file__))
    dump_obj = BestDump(folder, "111")

    test_pulse = np.array([1, 2, 3])
    test_timegrid = np.array([4, 5, 6])
    test_params = np.array([7, 8, 9])

    dump_obj.dump_controls([test_pulse], [test_timegrid], [test_params], False)

    outfile_path = os.path.join(dump_obj.best_controls_path, "111_best_controls.npz")
    file_exists = os.path.exists(outfile_path)
    assert not file_exists


def test_other_dumps():

    folder = os.path.dirname(os.path.realpath(__file__))
    dump_obj = BestDump(folder, "000")

    test_pulse = np.array([1, 2, 3])
    outfile_name = "my_test.txt"
    dump_obj.other_dumps(outfile_name, test_pulse)
    outfile_path = os.path.join(dump_obj.best_controls_path, outfile_name)
    test_file_load = np.loadtxt(outfile_path)
    # print(test_file_load)
    assert (test_file_load == test_pulse).all()
    os.remove(outfile_path)



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from abc import abstractmethod
from typing import Callable

import numpy as np

from quocslib.tools.randomgenerator import RandomNumberGenerator


class BasePulse:
    """
    This is the main class for Pulse. Every pulse has to inherit this class.
    """

    control_parameters_number: int
    optimized_control_parameters: np.ndarray
    final_time: float

    def __init__(self,
                 map_index=-1,
                 pulse_name="pulse",
                 bins_number=101,
                 time_name="time",
                 lower_limit=0.0,
                 upper_limit=1.0,
                 amplitude_variation=0.1,
                 initial_guess=None,
                 scaling_function=None,
                 is_shrinked: bool = False,
                 shaping_options: list = None,
                 overwrite_base_pulse: bool = False,
                 rng: RandomNumberGenerator = None,
                 **kwargs):
        """
        Here we defined all the basic features a pulse should have.

        :param int map_index: index number for pulse control parameters association
        :param str pulse_name: Pulse name
        :param int bins_number: Number of bins
        :param str time_name: Name of the time associated to this pulse
        :param float lower_limit: Lower amplitude limit of the pulse
        :param float upper_limit: Upper amplitude limit of the pulse
        :param float amplitude_variation: amplitude variation of the pulse
        :param dict initial_guess: dictionary with initial guess information
        :param dict scaling_function: dictionary with scaling function information
        :param kwargs: Other arguments
        """
        # The arguments did not use here, use for the other class
        # super().__init__(**kwargs)
        # Pulse name
        self.pulse_name = pulse_name
        # Bins number
        self.bins_number = bins_number
        # Base Pulse
        self.base_pulse = np.zeros(self.bins_number)
        # Time grid initialization
        self.time_grid = np.zeros(self.bins_number)
        # Time
        self.time_name = time_name
        # Amplitude Limits
        # TODO Implement check amplitude lower < amplitude upper
        (self.amplitude_lower, self.amplitude_upper) = (lower_limit, upper_limit)
        # Amplitude variation
        self.amplitude_variation = amplitude_variation
        # Create the parameter indexes list. It is used
        self.control_parameters_list = [map_index + i + 1 for i in range(self.control_parameters_number)]
        # Update the map_index number for the next pulse
        self.last_index = self.control_parameters_list[-1]
        # Shaping options
        print("Testing shaping option list mode")
        shaping_option_dict = {
            "add_base_pulse": self.add_base_pulse,
            "add_initial_guess": self.add_initial_guess,
            "limit_pulse": self.limit_pulse,
            "scale_pulse": self.scale_pulse
        }
        if shaping_options is None:
            self.shaping_options = [self.add_base_pulse, self.add_initial_guess, self.scale_pulse, self.limit_pulse]
        else:
            self.shaping_options = []
            for op_str in shaping_options:
                if op_str in shaping_option_dict:
                    self.shaping_options.append(shaping_option_dict[op_str])
                else:
                    print("Warning: {0} pulse option is not implemented".format(op_str))
        # Initial guess function
        function_type = initial_guess["function_type"]
        if function_type == "lambda_function":
            initial_guess_pulse = eval(initial_guess["lambda_function"])
        elif function_type == "list_function":
            initial_guess_pulse = np.asarray(initial_guess["list_function"])
        else:
            initial_guess_pulse = lambda t: 0.0 * t
        self.initial_guess_pulse = initial_guess_pulse
        # Scaling function
        function_type = scaling_function["function_type"]
        if function_type == "lambda_function":
            scaling_function = eval(scaling_function["lambda_function"])
        elif function_type == "list_function":
            scaling_function = np.asarray(scaling_function["list_function"])
        else:
            scaling_function = lambda t: 1.0 * t
        self.scaling_function = scaling_function
        # Shrink option
        self.is_shrinked = is_shrinked
        # Overwrite the base pulse at the end of the superiteration
        self.overwrite_base_pulse = overwrite_base_pulse
        # Random number generator
        self.rng = rng

    def set_control_parameters_list(self, map_index):
        """Set the control parameters list. It is used when the Chopped Basis changes during SIs"""
        self.control_parameters_list = [map_index + i + 1 for i in range(self.control_parameters_number)]

    def _set_time_grid(self, final_time: float) -> None:
        """Set the time grid"""
        self.final_time = final_time
        self.time_grid = np.linspace(0, final_time, self.bins_number)

    def _get_build_pulse(self) -> np.ndarray:
        """Build the pulse with all the constraints"""
        optimal_pulse = self._get_shaped_pulse()
        for op in self.shaping_options:
            optimal_pulse = op(optimal_pulse)
        # Pulse operations
        # Make a loop with all the operation that the users wants to apply to the pulse shaping
        # optimal_pulse_total = self._get_initial_guess() + self._constrained_pulse(optimal_pulse)
        return optimal_pulse

    def _constrained_pulse(self, optimal_pulse: np.ndarray) -> np.ndarray:
        """Apply further constraints to the final pulse"""
        # Shrink the optimal pulse
        optimal_limited_pulse = self._get_limited_pulse(optimal_pulse)
        optimal_scaled_pulse = optimal_limited_pulse * self._get_scaling_function()
        return optimal_scaled_pulse

    def _shrink_pulse(self, optimal_pulse: np.ndarray) -> np.ndarray:
        """Shrink the optimal pulse to respect the amplitude limits"""
        # Upper - lower bounds
        u_bound = self.amplitude_upper
        l_bound = self.amplitude_lower
        # Upper - lower values of the optimal pulse
        u_value = np.max(optimal_pulse)
        l_value = np.min(optimal_pulse)
        # Distances to the bounds
        u_distance = u_value - u_bound
        l_distance = l_bound - l_value
        # Check if the bounds are not respect by the optimal pulse
        if u_distance > 0.0 or l_distance > 0.0:
            # Calculate the middle position between the max and min amplitude
            distance_u_l_value = (u_value + l_value) / 2.0
            # Move the pulse to the center of the axis
            v_optimal_pulse = optimal_pulse - distance_u_l_value
            # Move the bounds to the center of the axis respect the optimal pulses
            # distance_u_l_bound = (u_bound + l_bound) / 2.0
            v_u_bound, v_l_bound = [u_bound - distance_u_l_value, l_bound - distance_u_l_value]
            # Check which is the greatest virtual distance
            v_u_value = np.max(v_optimal_pulse)
            v_l_value = np.min(v_optimal_pulse)
            # The distance is preserved under this transformation
            v_l_distance = l_distance
            v_u_distance = u_distance
            # Calculate the max distance and assign the max bound in the virtual frame
            if v_u_distance >= v_l_distance:
                max_value = v_u_value
                max_bound = v_u_bound
            else:
                max_value = v_l_value
                max_bound = v_l_bound
            # Calculate the shrink factor (< 1.0)
            shrink_factor = max_bound / max_value
            # rescale the virtual pulse
            shrinked_v_optimal_pulse = shrink_factor * v_optimal_pulse
            # Go back to the non virtual pulse, i.e. a transformation + distance_u_l_value
            shrinked_pulse = shrinked_v_optimal_pulse + distance_u_l_value
            # Re-assign the optimal pulse
            optimal_pulse = shrinked_pulse

        return optimal_pulse

        def _shrink_pulse_2(self, optimal_pulse: np.ndarray) -> np.ndarray:
            """Shrink the optimal pulse to respect the amplitude limits"""

            uiTotal = optimal_pulse
            lb = self.amplitude_lower
            ub = self.amplitude_upper
            ui = uiTotal.copy()
            temp_max = np.amax(ui)
            temp_min = np.amin(ui)

            # if the whole pulse is higher or lower than the limits take the limits
            if (temp_max >= ub and temp_min >= ub) or (temp_max <= lb and temp_min <= lb):
                ui = np.maximum(np.minimum(uiTotal, ub), lb)  # cut off the pulse

            # otherwise shrink the pulse if it exceeds the limits accordingly
            else:
                if temp_max > ub:
                    width_current = abs(temp_max - temp_min)
                    width_aim = abs(ub - temp_min)
                    correction = width_aim / width_current
                    # safety precaution just in case
                    if correction > 1:
                        correction = 1
                    # shift pulse down so that minimum is at zero, shrink, and move up again
                    ui = (ui - temp_min) * correction + np.full(len(ui), temp_min)

                temp_max = np.amax(ui)
                temp_min = np.amin(ui)

                if temp_min < lb:
                    width_current = abs(temp_max - temp_min)
                    width_aim = abs(temp_max - lb)
                    correction = width_aim / width_current
                    # safety precaution just in case
                    if correction > 1:
                        correction = 1

                    # shift pulse down so that maximum is at zero, shrink, and move up again
                    ui = (ui - temp_max) * correction + np.full(len(ui), temp_max)

        return ui

    def get_pulse(self, optimized_parameters_vector: np.ndarray, final_time: float = 1.0) -> np.ndarray:
        """Set the optimized control parameters, the time grid, and return the pulse"""
        self._set_control_parameters(optimized_parameters_vector)
        self._set_time_grid(final_time)
        return self._get_build_pulse()

    def set_base_pulse(self, optimized_control_parameters: np.ndarray, final_time: float = 1.0) -> None:
        """Set the base optimal pulse"""
        self._set_control_parameters(optimized_control_parameters)
        self._set_time_grid(final_time)
        if self.overwrite_base_pulse:
            self.base_pulse = self._get_shaped_pulse()
        else:
            self.base_pulse += self._get_shaped_pulse()

    def _set_control_parameters(self, optimized_control_parameters: np.ndarray) -> None:
        """Set the optimized control parameters vector"""
        # TODO Check if the optimized control parameters vector has a size equal to the control parameters number
        #  of this pulse, otherwise raise an error
        self.optimized_control_parameters = optimized_control_parameters

    def _get_scaling_function(self) -> np.ndarray:
        """Get the array scaling function"""
        if isinstance(self.scaling_function, Callable):
            scaling_function_t = self.scaling_function(self.time_grid)
        elif isinstance(self.scaling_function, np.ndarray):
            scaling_function_t = self.scaling_function
        else:
            # TODO Handle here
            print("Warning: scaling function not good. Do with 1.0")
            scaling_function_t = (lambda t: 1.0)(self.time_grid)
        return scaling_function_t

    def _get_initial_guess(self) -> np.ndarray:
        """Get the initial guess pulse"""
        if isinstance(self.initial_guess_pulse, Callable):
            initial_guess_t = self.initial_guess_pulse(self.time_grid)
        elif isinstance(self.initial_guess_pulse, np.ndarray):
            initial_guess_t = self.initial_guess_pulse
        else:
            # TODO Handle here
            print("Warning: initial guess function not good. Do with 0.0")
            initial_guess_t = (lambda t: 0.0)(self.time_grid)
        return self._get_limited_pulse(initial_guess_t)

    def add_base_pulse(self, pulse: np.ndarray):
        """Add the base pulse"""
        return pulse + self.base_pulse

    def add_initial_guess(self, pulse: np.ndarray):
        """Add the initial pulse to the optimal pulse"""
        return self._get_initial_guess() + pulse

    def limit_pulse(self, pulse: np.ndarray):
        """Apply the constraints to the optimal pulse"""
        return self._get_limited_pulse(pulse)

    def scale_pulse(self, pulse: np.ndarray):
        """Scale the optimal pulse accordingly to the scaling function"""
        return self._get_scaling_function() * pulse

    def _get_limited_pulse(self, optimal_pulse: np.ndarray):
        """Cut the pulse with the amplitude limits constraints"""
        if self.is_shrinked:
            return self._shrink_pulse(optimal_pulse)
        else:
            return np.maximum(np.minimum(optimal_pulse, self.amplitude_upper), self.amplitude_lower)

    @abstractmethod
    def _get_shaped_pulse(self) -> np.ndarray:
        """Just an abstract method to get the optimized shape pulse"""



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from abc import ABCMeta, abstractmethod


class AbstractFoM(metaclass=ABCMeta):
    """Abstract class for figure of merit evaluation"""
    @abstractmethod
    def get_FoM(self, pulses_list, time_grids_list, parameters_list) -> dict:
        """Abstract method for figure of merit evaluation. It returns a dictionary with
        the FoM key inside"""
